# Potential micro-optimizations using sse2 and other tricks.

These are all improving CPU time. The effect on overall
performance might not be noticeable. Especially if the operation is
I/O bound.

* Improve range() (already sse2 enabled) to handle more cases.
  Note that range() is used only when writing data.
  - Support int8 and int16 samples.
  - Support buffers that include padding.
  - Efficient handling also for buffers not aligned to 16 bytes.
  - The Inf/Nan test can be done with fewer instructions.

* Improve IsFiniteT() and IsNanT() by using bit fiddling. Used among
  other places in the WeightedAverage LOD algorithm. If range() hadn't
  been sse2 enabled then it would also have fixed that performance,
  albeit not quite as much as sse did. This tweak is done. It has
  been shown to have a positive effect for IsFiniteT(float) both on
  the gnu and even more on the msvc compiler. I also have similar code
  for IsFiniteT(double), IsNanT(float), and IsNanT(double) but I
  won't enable those unless I take the time to measure performance.
  Caveat: Whether this speeds up or slows down computation is highly
  compiler dependent. So we might need to test the speed each time we
  switch compilers.

* The OpenMP loop running RoundAndClip for 256*1024 samples might be
  better off as single threaded. The overhead might be too high. This
  is very dependent on the environment, though.

* Implement a ScaleRoundAndClip as a vector version of RoundAndClip.
  Set the per-thread rounding mode to nearest for the duration of the
  loop. This saves a branch when rounding: return (int)d; instead of
  (d>0) ? (int)(d+0.5) : (int)(d-0.5); Caveat: Measurements using
  similar code show that this actually runs slower on g++. Which makes
  no sense at all. Possibly the g++ optimizer just got too clever. I
  haven't tested in windows. The function only applies to writing
  float data to a file that has int8 or int16 storage.

* Implement ScaleRoundAndClip using sse2. This will probably speed up
  writing float data to a file that has int8 or int16 storage.
  Especially if we need to fix and slow down the existing
  RoundAndClip. Technically, today's implementation triggers undefined
  behavior. One counter-indication is that s_scaleFromFloat() runs
  multi-threaded. Which already makes it reasonably fast. Another
  issue is that this tweak, just like range(), is limited to certain
  scenarios.

* Implement the entire WeightedAverage LOD algorithm using sse2.
  Separate implementataions needed for int8, int16, and float. This
  helps writing speed only. And, like s_scaleFromFloat() this uses
  multi-threading. So it isn't a huge bottleneck today.

* Implement histogram- and statistics computation using sse2. Looks
  like a fair amount of work to implement without cutting any corners.
  And less efficient because it will only handle two samples in
  parallel. We won't know the actual improvement until the
  implementation is complete.

* Implement some or all of the above using sse4, avx, or whatever
  other technology is available. This will probably be faster than
  sse2. But it will not work on all processors. This means we need to
  keep the sse2 code as a fallback. Testing becomes a nightmare.

### Technical information

Why isn't std::isfinite optimal?

g++ -O3 generates the following code in g++ v8.5:

std::isfinite() and __builtin_isfinite() generate identical code, using sse2 without being asked to.

```
    andps	.LC0(%rip), %xmm0
    movss	.LC1(%rip), %xmm1
    ucomiss	%xmm0, %xmm1
    setnb	%al
    ret
.LC0:
    .long	2147483647, 0, 0, 0
.LC1:
    .long	2139095039
```

The hand-coded bit fiddling in IsFiniteT ends up in cpu registers.

```
    movd	%xmm0, %eax
    andl	$2147483647, %eax
    cmpl	$2139095039, %eax
    setbe	%al
    ret
```

In gcc 8.5 the CPU version of isfinite is significantly faster. This behavior is very compiler dependent. Caveat: This code was generated by a simple test program. The real function has the IsFiniteT code inlined into a larger function. It might look different there.

std::isfinite will use an actual function call on the current Windows (msvc) compiler. Even when optimized. This means that the bit fiddling is even more useful there. For now.
