// Copyright 2017-2020, Schlumberger
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

//Adapted from: Zgy/UtilityLib/StatisticData

#include "statisticdata.h"

#include <cstdint>
#include <limits>
#include <assert.h>
#include <algorithm>
#include <sstream>

namespace InternalZGY {
#if 0
}
#endif

/**
 * Initialize a StatisticData to empty.
 * min/max are by definition irrelevant when count is zero,
 * but they will be set to 0 here just to have a consistent value.
 */
StatisticData::StatisticData()
  : cnt_(0), inf_(0), sum_(0), ssq_(0)
  , min_(std::numeric_limits<double>::infinity())
  , max_(-std::numeric_limits<double>::infinity())
{
}

/**
 * Initialize a StatisticData from discrete values.
 */
StatisticData::StatisticData(count_type cnt, count_type inf, double sum, double ssq, double min, double max)
  : cnt_(cnt), inf_(inf), sum_(sum), ssq_(ssq), min_(min), max_(max)
{
}

/**
 * Create a StatisticData from histogram information.
 *
 * If is_int8 is true, the histogram was populated from 8-bit data
 * and the histogram has 256 bins. In this case the result will be
 * precise. The method needs to be told about this. Otherwise it
 * would have to add 1/2 bin width slop, not knowing that there can
 * only be a single value in each bin. That value might not be
 * integral any longer due to scaling, buit there can still only be one.

 * If is_int8 is false, this isn't as accurate as the actual statistics
 * since each sample will be rounded to the center of the bin it falls in.
 * Also, Inf will always be 0 since the histogram doesn't count outliers.
 * The result is still useful as a consistency check. And also to update
 * the min/max range in the real statistics after subtracting samples.
 *
 * Used by HistogramBuilder::fastAddSignedChar256() which provides a
 * not insignificant performance boost in int8 files.
 *
 * Used by HistogramBuilder::gethiststats(), but that method itself is no
 * longer invoked because it gives incorrect results when the histogram
 * range cannot grow.
 */
StatisticData::StatisticData(const count_type *bins, int nbins, double range_min, double range_max, bool is_int8)
: cnt_(0), inf_(0), sum_(0), ssq_(0)
, min_(std::numeric_limits<double>::infinity())
, max_(-std::numeric_limits<double>::infinity())
{
  const double begin = range_min;
  const double width = (range_max - range_min) / (nbins - 1);
  const double slop = is_int8 ? 0 : width/2;

  for (int binno = 0; binno < nbins; ++binno) {
    count_type count = bins[binno];
    double value = begin + binno*width;
    if (count != 0) {
      if (cnt_ == 0)
        min_ = value - slop; // first non-empty bin found
      max_ = value + slop; // last (so far) non-empty bin found.
    }
    cnt_ += count;
    sum_ += count * value;
    ssq_ += count * value * value;
  }
}

/**
 * \brief Add more samples to an existing StatisticData.
 *
 * \detailed
 *
 * other is also allowed to hold negative counts,
 * this will cause samples to be removed. But the
 * min/max range will still be expanded.
 *
 * See the class header for a long explanation.
 *
 */
StatisticData& StatisticData::operator+=(const StatisticData& other)
{
  if (other.min_ <= other.max_) {
    if (min_ <= max_) {
      // Both non-empty.
      min_ = std::min(min_, other.min_);
      max_ = std::max(max_, other.max_);
    }
    else {
      // *this is empty, other is not.
      min_ = other.min_;
      max_ = other.max_;
    }
  }
  cnt_ += other.cnt_;
  sum_ += other.sum_;
  ssq_ += other.ssq_;
  inf_ += other.inf_; // If other is empty, only inf_ can change.
  return *this;
}

/**
 * \brief Remove samples from an existing StatisticData.
 *
 * \detailed
 *
 * The resulting min/max range will be the union of the two classes,
 * not the intersection. Could also have chosen to leave the
 * min/max uncanged. But it is more consistent to do a union,
 * because -a+b will then work like b-a.  Either way, you can call
 * trimRange() afterwards if you know what the range ought to be.
 *
 * See the class header for a long explanation.
 */
StatisticData& StatisticData::operator-=(const StatisticData& other)
{
  if (other.min_ <= other.max_) {
    if (min_ <= max_) {
      min_ = std::min(min_, other.min_);
      max_ = std::max(max_, other.max_);
    }
    else {
      min_ = other.min_;
      max_ = other.max_;
    }
  }
  cnt_ -= other.cnt_;
  sum_ -= other.sum_;
  ssq_ -= other.ssq_;
  inf_ -= other.inf_;
  return *this;
}

/**
 * Multiply StatisticData with a constant N, equivalent to creating
 * a new instance and adding the old one to it N times. N can also be
 * negative. The min/max range is not affected.
 */
StatisticData& StatisticData::operator*=(count_type factor)
{
  cnt_ *= factor;
  inf_ *= factor;
  sum_ *= factor;
  ssq_ *= factor;
  return *this;
}

/**
 * Try to narrow the range found in the statistics based on the
 * information we have in the histogram. This is particularly
 * useful after subtracting something from the histogram.
 *
 * The range found in the histogram is normally wider than or equal to
 * the statistics range. If this is not the case, we know the min/max
 * in the statistics are incorrect.  We cannot know the precise new
 * range without rescanning all the data, but we can approximate it
 * using the histogram.  There is no point in doing any of this if the
 * histogram is empty. In that case min/max are irrelevant, and will
 * be reset once data has been added.  If any of the overflow bins are
 * in use, these are ignored.  That should not happen in normal
 * circumstances, but could occur if stats are calculated from
 * application provided data, but maintained in a fixed histogram
 * because that is what the file format dictates.
 */
void StatisticData::trimRange(const count_type *bins, int nbins, double range_min, double range_max)
{
  if (cnt_ != 0) {
    const double begin = range_min;
    const double width = (range_max - range_min) / (nbins - 1);
    const double slop = width/2;

    int lobin = 0;
    while (bins[lobin] == 0 && lobin < nbins)
      ++lobin;
    int hibin = nbins - 1;
    while (bins[hibin] == 0 && hibin >= 0)
      --hibin;
    if (lobin <= hibin) {
      // test is needed because all bins might have been empty,
      // due to an inconsistency or because all the values were
      // in the overflow bins or NaN (currently not counted).
      // Not much we can do if that is the case.
      // Applying the transform gives the center of the bin,
      // the actual range of the data can be 1/2 bin more at either end.
      double histmin = begin + width*lobin - slop;
      double histmax = begin + width*hibin + slop;
      if (min_ < histmin)
        min_ = histmin;
      if (max_ > histmax)
        max_ = histmax;
    }
  }
}

std::string
StatisticData::toString() const
{
  std::stringstream ss;
  if (getcnt() == 0 && getsum() == 0 && getssq() == 0 && getinf() == 0 && getmin() > getmax()) {
    ss << "empty";
  }
  else if (getcnt() > 0) {
    ss << "cnt " << getcnt() << " inf " << getinf()
       << " min " << getmin() << " max " << getmax()
       << " avg " << float(getsum() / getcnt())
       << " rms " << float(getssq() / getcnt());
  }
  else {
    ss << "cnt " << getcnt() << " inf " << getinf()
       << " min " << getmin() << " max " << getmax()
       << " sum " << float(getsum())
       << " rms " << float(getssq());
  }
  return ss.str();
}

} // end namespace
